/*
 * Part of Very Secure FTPd
 * Licence: GPL v2
 * Author: Dmitriy Balashov
 * charconv.c
 */

#include "charconv.h"
#include "tunables.h"
#include "session.h"
#include "str.h"
#include "sysutil.h"

/* Tables mapping supported codepage names to runtime variables */
static struct available_charsets
{
  const char* p_charset_name;
  int p_variable;
}
available_charsets_array[] =
{
  { VSFTP_CP_NONE   , VSFTP_C_NONE },
  { VSFTP_CP_UTF8   , VSFTP_C_UTF8 },
  { VSFTP_CP_WIN1251, VSFTP_C_1251 },
  { VSFTP_CP_KOI8R  , VSFTP_C_878  },
  { VSFTP_CP_DOS    , VSFTP_C_866  },
  { VSFTP_CP_1251   , VSFTP_C_1251 },
  { VSFTP_CP_878    , VSFTP_C_878  },
  { VSFTP_CP_866    , VSFTP_C_866  },
  { 0, 0 }
};

/* Available convertions */
static struct available_convertions
{
  int source;
  int destination;
}
available_convertions_array[] =
{
  { VSFTP_C_UTF8, VSFTP_C_1251 },
  { VSFTP_C_UTF8, VSFTP_C_878  },
  { VSFTP_C_UTF8, VSFTP_C_866  },
  { VSFTP_C_1251, VSFTP_C_UTF8 },
  { VSFTP_C_1251, VSFTP_C_878  },
  { VSFTP_C_1251, VSFTP_C_866  },
  { VSFTP_C_878 , VSFTP_C_UTF8 },
  { VSFTP_C_878 , VSFTP_C_1251 },
  { VSFTP_C_878 , VSFTP_C_866  },
  { VSFTP_C_866 , VSFTP_C_UTF8 },
  { VSFTP_C_866 , VSFTP_C_1251 },
  { VSFTP_C_866 , VSFTP_C_878  },
  { 0, 0 }
};

/* Tables mapping codepage */
struct codepage_map codepage_utf8_array[] =
{
  {2, 0},
#ifdef UTF8_LOW_FIRST
  {0x90D0,  1}, {0x91D0,  2}, {0x92D0,  3}, {0x93D0,  4}, {0x94D0,  5}, {0x95D0,  6}, {0x81D0,  7}, {0x96D0,  8}, {0x97D0,  9},
  {0x98D0, 10}, {0x99D0, 11}, {0x9AD0, 12}, {0x9BD0, 13}, {0x9CD0, 14}, {0x9DD0, 15}, {0x9ED0, 16}, {0x9FD0, 17},
  {0xA0D0, 18}, {0xA1D0, 19}, {0xA2D0, 20}, {0xA3D0, 21}, {0xA4D0, 22}, {0xA5D0, 23}, {0xA6D0, 24}, {0xA7D0, 25},
  {0xA8D0, 26}, {0xA9D0, 27}, {0xAAD0, 28}, {0xABD0, 29}, {0xACD0, 30}, {0xADD0, 31}, {0xAED0, 32}, {0xAFD0, 33},
  {0xB0D0, 34}, {0xB1D0, 35}, {0xB2D0, 36}, {0xB3D0, 37}, {0xB4D0, 38}, {0xB5D0, 39}, {0x91D1, 40}, {0xB6D0, 41}, {0xB7D0, 42},
  {0xB8D0, 43}, {0xB9D0, 44}, {0xBAD0, 45}, {0xBBD0, 46}, {0xBCD0, 47}, {0xBDD0, 48}, {0xBED0, 49}, {0xBFD0, 50},
  {0x80D1, 51}, {0x81D1, 52}, {0x82D1, 53}, {0x83D1, 54}, {0x84D1, 55}, {0x85D1, 56}, {0x86D1, 57}, {0x87D1, 58},
  {0x88D1, 59}, {0x89D1, 60}, {0x8AD1, 61}, {0x8BD1, 62}, {0x8CD1, 63}, {0x8DD1, 64}, {0x8ED1, 65}, {0x8FD1, 66},
#else
  {0xD090,  1}, {0xD091,  2}, {0xD092,  3}, {0xD093,  4}, {0xD094,  5}, {0xD095,  6}, {0xD081,  7}, {0xD096,  8}, {0xD097,  9},
  {0xD098, 10}, {0xD099, 11}, {0xD09A, 12}, {0xD09B, 13}, {0xD09C, 14}, {0xD09D, 15}, {0xD09E, 16}, {0xD09F, 17},
  {0xD0A0, 18}, {0xD0A1, 19}, {0xD0A2, 20}, {0xD0A3, 21}, {0xD0A4, 22}, {0xD0A5, 23}, {0xD0A6, 24}, {0xD0A7, 25},
  {0xD0A8, 26}, {0xD0A9, 27}, {0xD0AA, 28}, {0xD0AB, 29}, {0xD0AC, 30}, {0xD0AD, 31}, {0xD0AE, 32}, {0xD0AF, 33},
  {0xD0B0, 34}, {0xD0B1, 35}, {0xD0B2, 36}, {0xD0B3, 37}, {0xD0B4, 38}, {0xD0B5, 39}, {0xD191, 40}, {0xD0B6, 41}, {0xD0B7, 42},
  {0xD0B8, 43}, {0xD0B9, 44}, {0xD0BA, 45}, {0xD0BB, 46}, {0xD0BC, 47}, {0xD0BD, 48}, {0xD0BE, 49}, {0xD0BF, 50},
  {0xD180, 51}, {0xD181, 52}, {0xD182, 53}, {0xD183, 54}, {0xD184, 55}, {0xD185, 56}, {0xD186, 57}, {0xD187, 58},
  {0xD188, 59}, {0xD189, 60}, {0xD18A, 61}, {0xD18B, 62}, {0xD18C, 63}, {0xD18D, 64}, {0xD18E, 65}, {0xD18F, 66},
#endif
  {0, 0}
};

struct codepage_map codepage_1251_array[] =
{
  {1, 0},
  {0xC0,  1}, {0xC1,  2}, {0xC2,  3}, {0xC3,  4}, {0xC4,  5}, {0xC5,  6}, {0xA8,  7}, {0xC6,  8}, {0xC7,  9},
  {0xC8, 10}, {0xC9, 11}, {0xCA, 12}, {0xCB, 13}, {0xCC, 14}, {0xCD, 15}, {0xCE, 16}, {0xCF, 17},
  {0xD0, 18}, {0xD1, 19}, {0xD2, 20}, {0xD3, 21}, {0xD4, 22}, {0xD5, 23}, {0xD6, 24}, {0xD7, 25},
  {0xD8, 26}, {0xD9, 27}, {0xDA, 28}, {0xDB, 29}, {0xDC, 30}, {0xDD, 31}, {0xDE, 32}, {0xDF, 33},
  {0xE0, 34}, {0xE1, 35}, {0xE2, 36}, {0xE3, 37}, {0xE4, 38}, {0xE5, 39}, {0xB8, 40}, {0xE6, 41}, {0xE7, 42},
  {0xE8, 43}, {0xE9, 44}, {0xEA, 45}, {0xEB, 46}, {0xEC, 47}, {0xED, 48}, {0xEE, 49}, {0xEF, 50},
  {0xF0, 51}, {0xF1, 52}, {0xF2, 53}, {0xF3, 54}, {0xF4, 55}, {0xF5, 56}, {0xF6, 57}, {0xF7, 58},
  {0xF8, 59}, {0xF9, 60}, {0xFA, 61}, {0xFB, 62}, {0xFC, 63}, {0xFD, 64}, {0xFE, 65}, {0xFF, 66},
  {0, 0}
};

struct codepage_map codepage_878_array[] = // may be available error
{
  {1, 0},
  {225,  1}, {226,  2}, {247,  3}, {231,  4}, {228,  5}, {229,  6}, {179,  7}, {246,  8}, {250,  9},
  {233, 10}, {234, 11}, {235, 12}, {236, 13}, {237, 14}, {238, 15}, {239, 16}, {240, 17},
  {242, 18}, {243, 19}, {244, 20}, {245, 21}, {230, 22}, {232, 23}, {227, 24}, {254, 25},
  {251, 26}, {253, 27}, {154, 28}, {249, 29}, {248, 30}, {252, 31}, {224, 32}, {241, 33},
  {193, 34}, {194, 35}, {215, 36}, {199, 37}, {196, 38}, {197, 39}, {163, 40}, {214, 41}, {218, 42},
  {201, 43}, {202, 44}, {203, 45}, {204, 46}, {205, 47}, {206, 48}, {207, 49}, {208, 50},
  {210, 51}, {211, 52}, {212, 53}, {213, 54}, {198, 55}, {200, 56}, {195, 57}, {222, 58},
  {219, 59}, {221, 60}, {223, 61}, {217, 62}, {216, 63}, {220, 64}, {192, 65}, {209, 66},
  {0, 0}
};

struct codepage_map codepage_866_array[] =
{
  {1, 0},
  {0x80,  1}, {0x81,  2}, {0x82,  3}, {0x83,  4}, {0x84,  5}, {0x85,  6}, {0xF0,  7}, {0x86,  8}, {0x87,  9},
  {0x88, 10}, {0x89, 11}, {0x8A, 12}, {0x8B, 13}, {0x8C, 14}, {0x8D, 15}, {0x8E, 16}, {0x8F, 17},
  {0x90, 18}, {0x91, 19}, {0x92, 20}, {0x93, 21}, {0x94, 22}, {0x95, 23}, {0x96, 24}, {0x97, 25},
  {0x98, 26}, {0x99, 27}, {0x9A, 28}, {0x9B, 29}, {0x9C, 30}, {0x9D, 31}, {0x9E, 32}, {0x9F, 33},
  {0xA0, 34}, {0xA1, 35}, {0xA2, 36}, {0xA3, 37}, {0xA4, 38}, {0xA5, 39}, {0xF1, 40}, {0xA6, 41}, {0xA7, 42},
  {0xA8, 43}, {0xA9, 44}, {0xAA, 45}, {0xAB, 46}, {0xAC, 47}, {0xAD, 48}, {0xAE, 49}, {0xAF, 50},
  {0xE0, 51}, {0xE1, 52}, {0xE2, 53}, {0xE3, 54}, {0xE4, 55}, {0xE5, 56}, {0xE6, 57}, {0xE7, 58},
  {0xE8, 59}, {0xE9, 60}, {0xEA, 61}, {0xEB, 62}, {0xEC, 63}, {0xED, 64}, {0xEE, 65}, {0xEF, 66},
  {0, 0}
};

map_ptr map_array[] =
{
  0, codepage_utf8_array, codepage_1251_array, codepage_878_array, codepage_866_array
};

const char* vsf_charset_name(int code)
{
  return available_charsets_array [code].p_charset_name;
}

int vsf_charset_init(const char* p_str)
{
  const struct available_charsets* charsets = available_charsets_array;
  
  while (charsets->p_charset_name != 0)
  {
    if (str_equal_str (charsets->p_charset_name, p_str))
    {
      return charsets->p_variable;
    }
    charsets++;
  }
  
  return 0;
}

int vsf_available_convertion(int srcCode, int dstCode)
{
  const struct available_convertions* aconv = available_convertions_array;
  
  while (aconv->source != 0)
  {
    if (srcCode == aconv->source && dstCode == aconv->destination) return 1;
    aconv++;
  }

  return 0;
}

void vsf_charset_conv(struct vsf_session* p_sess, struct mystr* p_str, int direction)
{
  const char* srcbuf;
  unsigned int srclen;
  char* dstbuf;
  map_ptr src, dst;
  unsigned int s2bytes, d2bytes, t2;
  unsigned int srcpos = 0, dstpos = 0;
  unsigned int chpos;
  unsigned int char_code = 0;

  if (!p_sess->enable_convertion || !p_sess->remote_charset || !tunable_local_codepage) return;

  srclen = str_getlen(p_str);     // Len of source string
  srcbuf = str_getbuf(p_str);

  if (direction == 1)
  {
    src = map_array [p_sess->remote_charset];
    dst = map_array [tunable_local_codepage];
  }
  else
  {
    src = map_array [tunable_local_codepage];
    dst = map_array [p_sess->remote_charset];
  }

  s2bytes = src [0].char_code;    // Getting char size with charcode over 127
  d2bytes = dst [0].char_code;
  
  dstbuf = vsf_sysutil_malloc(srclen * d2bytes + d2bytes); // Segmentation Failure

  while (srcpos < srclen)
  {
    char_code = (unsigned char)srcbuf [srcpos];
    if (char_code > 127)
    {
      if (s2bytes == 2)
      {
        srcpos++;
#ifdef UTF8_LOW_FIRST
        char_code = char_code | ((unsigned char)srcbuf [srcpos] << 8);
#else
        char_code = (char_code << 8) | (unsigned char)srcbuf [srcpos];
#endif
      }
      
      chpos = 1;
      while (src [chpos].char_code && src [chpos].char_code != char_code)
      {
        chpos++;
      }
      if (src [chpos].char_code)
      {
        char_code = dst [src [chpos].conv_code].char_code;
        t2 = d2bytes;
      }
      else
      {
        t2 = s2bytes;
      }

#ifdef UTF8_LOW_FIRST
      dstbuf [dstpos] = (char)(char_code & 0xFF);
      if (t2 == 2)
      {
        dstpos++;
        dstbuf [dstpos] = (char)((char_code >> 8) & 0xFF);
      }
#else
      dstbuf [dstpos] = (char)((char_code >> 8) & 0xFF);
      if (t2 == 2)
      {
        dstpos++;
        dstbuf [dstpos] = (char)(char_code & 0xFF);
      }
#endif

    }
    else
    {
      dstbuf [dstpos] = srcbuf [srcpos];
    }
    srcpos++;
    dstpos++;
  }

  dstbuf [dstpos] = '\0';

  str_empty(p_str);
  str_append_text(p_str, dstbuf);

  vsf_sysutil_free(dstbuf);
}
